package grobs;

import java.awt.* ;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.* ;

import javax.swing.* ;

/**
 * <p>Title: Turtle World</p>
 * <p>Description: Turtle graphics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 */

public class GrobArena extends JComponent implements Iterable<Grob> {

    private java.util.List<Grob> list = new ArrayList<Grob>() ;
    private ArenaListener listener ;

    private Observer observer = new Observer() {
        public void update( Observable grob, Object arg ) {
            repaint() ; } } ;

    public GrobArena() {
        setPreferredSize( new Dimension(300,180)) ;
        setMinimumSize( new Dimension( 300, 180 ) ) ;
        setBackground(Color.WHITE); 
        ArenaMouseListener mouseListener = new ArenaMouseListener() ; 
        addMouseListener( mouseListener ) ;
        addMouseMotionListener( mouseListener ) ;
    }
    
    public void setListener( ArenaListener listener ) {
    	this.listener = listener ; }

    public void add( Grob grob ) {
        if( ! list.contains(grob) ) {
            list.add(grob) ;
            Collections.sort(list, new Comparator<Grob>(){
				@Override public int compare(Grob o1, Grob o2) {
					return o1.zIndex - o2.zIndex;
				}} ) ;
            grob.setArena(this);
            grob.addObserver( observer ) ; }
        repaint(); }

    public void remove( Grob grob ) {
        if( list.contains( grob ) ) {
            list.remove( grob ) ;
            grob.setArena(null);
            grob.deleteObserver( observer ); }
        repaint() ; }

    public void paintComponent( Graphics g ) {
        Graphics2D g2 = (Graphics2D)( g.create() ) ;
        g2.setBackground(getBackground()) ;
        g2.clearRect(0, 0, getWidth(), getHeight()) ;
        Iterator<Grob> it = list.iterator() ;
        while( it.hasNext() ) {
            Grob grob = it.next() ;
            grob.paint( g2 ) ; }
        g2.dispose() ;
    }

    public Iterator<Grob> iterator() {
        return list.iterator() ; }
    
    class ArenaMouseListener extends  MouseAdapter {
    	Grob draggedGrob = null ;
    	int lastX, lastY ;
    	
    	@Override
    	public void mousePressed(MouseEvent e) {
    		super.mousePressed(e);
    		Grob oldDraggedGrob = draggedGrob ;
    		lastX = e.getX() ;
    		lastY = e.getY() ;
    		listener.mousePressed(find(false, false)) ;
    	}
    	
    	@Override public void mouseReleased(MouseEvent e) {
    		super.mouseReleased(e);
    		//System.out.println("mouse released") ;
    		lastX = e.getX() ;
    		lastY = e.getY() ;
    		listener.mouseReleased(find(false, false)) ;
    		if( draggedGrob != null ) {
    			Grob target = find(false, true) ;
    			listener.endDrag( draggedGrob, target ) ; 
    			draggedGrob = null ;}
    	}
    	
    	@Override public void mouseDragged(MouseEvent e) {
    		super.mouseDragged(e);
    		//System.out.println("mouse dragged") ;
    		if( draggedGrob == null ) {
    			draggedGrob = find(true, false);
    			if( draggedGrob != null) {
    				listener.startDrag( draggedGrob ) ; } }
    		if( draggedGrob != null) {
    			int x = e.getX(), y = e.getY() ;
    			if( x == lastX && y == lastY ) return ;
    			draggedGrob.moveBy(x-lastX, y-lastY) ;
    			lastX = x ; lastY = y ; }
    	}

		private Grob find(boolean mustBeDraggable, boolean mustBeDropTarget) {
			Point2D point = new Point2D.Float(lastX, lastY ) ;
			Grob result = null ;
			for( Grob g : GrobArena.this ) {
				if( (mustBeDraggable && g.isDraggable() || mustBeDropTarget && g.isDropTarget())
				&& g.getShape().contains(point) ) {
					result = g ; } }
			return result ;
		}
    }
}