package grobs;

import java.awt.* ;
import java.awt.geom.* ;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.WritableRaster;
import java.net.URL;
import java.util.* ;

import javax.swing.ImageIcon;

/**
 * <p>Title: Turtle World</p>
 * <p>Description: Turtle graphics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Theodore Norvell
 * @version 1.0
 */

public class Turtle extends DynamicGrob {

    private double radarBeamAngle = 90.0 ;
    private double radarRadius = 0.0 ;
    private double radarDirection = 0.0 ;
    private java.util.List<Segment> trail = new LinkedList<Segment>() ;
    private double trailLength = 0.0 ;
    private double maxTrailLength = Double.POSITIVE_INFINITY ;
    private Segment lastSegment = null ; // Pen is down iff lastSegment != null
    private static final double FUZZ = 0.01 ;
    private double radius = 10 ;
    static private final ImageIcon originalmageIcon ;
    static private final double iconDeltaX = 16, iconDeltaY = 12.5;
    static {
    	URL url = Turtle.class.getResource("JesssTurtle0.png") ;
    	if( url == null ) {
    		System.out.println("URL for image is null") ;
    		originalmageIcon = null ;
    	} else {
    		originalmageIcon = new ImageIcon(url) ; }
    }
    private BufferedImage image ;
    
    public Turtle(Color c) {
    	setFillColor( c ) ;
    }

    public void paint( Graphics2D g ) {
    	monitor.enter() ;
    	try {
        int x = (int) state.position.x ;
        int y = (int) state.position.y ;

        double radians = orientationInRadians() ;
        // Radar
            Shape radar = getRadarShape() ;
            g.setColor( Color.pink ) ;
            g.fill( radar ) ;
            if( image == null ) {
            	//Head
            	int headRadius = 4 ;
            	double headOffset = state.speed==0.0 && state.angularVelocity==0.0
            			? radius - headRadius/2
            		    : radius + headRadius/2 ;
            	int headX = x + (int)( headOffset*Math.cos( radians ) ) ;
            	int headY = y + (int)( headOffset*Math.sin( radians ) ) ;
            	g.setColor( Color.black ) ;
            	g.fillOval(headX-headRadius, headY-headRadius, 2*headRadius, 2*headRadius);

            	// Body
            	g.setColor( fillColor );
            	g.fillOval(x-(int)radius, y-(int)radius, 2*(int)radius, 2*(int)radius);
            } else {
            	AffineTransform xform = AffineTransform.getTranslateInstance(x, y) ;
            	AffineTransform rotate = AffineTransform.getRotateInstance(radians) ;
            	AffineTransform shrink = AffineTransform.getScaleInstance(radius/16.0, radius/13.0) ;
            	AffineTransform centre = AffineTransform.getTranslateInstance(-iconDeltaX, -iconDeltaY ) ;
            	xform.concatenate(rotate);
            	xform.concatenate(shrink);
            	xform.concatenate(centre);
            	g.drawImage(image, xform, null) ;
            }

        // Trail
            for( int i=0 ; i<trail.size() ; ++i ) {
                Segment seg = trail.get(i) ;
                g.setColor( seg.color ) ;
                g.drawLine((int)seg.startPoint.x, (int)seg.startPoint.y,
                           (int)seg.endPoint.x, (int)seg.endPoint.y ) ; }
            g.setColor( Color.black ) ;
    } finally {
    	monitor.leave() ; }
    }
    
    @Override
    public void setFillColor( Color c ) {
    	super.setFillColor( c ) ;
    	recolorTurtle( c ) ;
    }
    
    private void recolorTurtle( Color c ) {
    	if( originalmageIcon == null ) {
    		image = null ; }
    	else {
    		image = deepCopy( originalmageIcon.getImage() ) ;
    		if( image == null ) return ;
    		float [] hsbNew = new float[3] ;
			int  newRGB = c.getRGB() ;
			//System.out.format("New colour %x", newRGB ); System.out.println();
			int newRed =   ((newRGB >> 16) & 0xFF) ;
			int newGreen = ((newRGB >>  8) & 0xFF) ;
			int newBlue =  ((newRGB      ) & 0xFF) ;
    		Color.RGBtoHSB(newRed, newGreen, newBlue, hsbNew) ;
    		float[] hsb = new float[3] ;
    		for( int i = 0, w = image.getWidth() ; i < w ; ++i ) {
    			for( int j = 0, h = image.getHeight() ; j < h ; ++j ) {
    				int rgb = image.getRGB(i, j); 
					//System.out.format("old colour %x", rgb ); System.out.println();
    				int alpha = (rgb >> 24) & 0xFF;
    				int red =   (rgb >> 16) & 0xFF;
    				int green = (rgb >>  8) & 0xFF;
    				int blue =  (rgb      ) & 0xFF;
    				if( alpha > 0 ) {
    					Color.RGBtoHSB(red, green, blue, hsb) ;
    					int rbgNew = Color.HSBtoRGB( hsbNew[0], hsbNew[1], hsb[2] ) ;
    					newRGB = (alpha << 24) | rbgNew & 0x00FFFFFF ;
    					//System.out.format("New colour %x", newRGB ); System.out.println();
    					image.setRGB(i, j, newRGB) ; } } } }
    }
    
    /** Source http://stackoverflow.com/questions/8864275/how-to-clone-image */
    static BufferedImage deepCopy( Image originalImage) {
    	int w = originalImage.getWidth(null) ;
    	int h = originalImage.getHeight(null) ;
    	if( w < 0 || h < 0 ) return null ;
    	BufferedImage copy = 
    			   new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
    	Graphics g = copy.createGraphics();
    	g.drawImage(originalImage, 0, 0, null);
    	g.dispose() ;
    	return copy ;
    }

    public Shape getShape() {
        /** @todo This should really be protected by the monitor,
         * as it accesses the state. But I can't do that
         * until the monitors are made reentrant */
        int x = (int) state.position.x ;
        int y = (int) state.position.y ;
        Shape s = new java.awt.geom.Ellipse2D.Double(x-radius, y-radius, 2*radius, 2*radius ) ;
        //Shape s = new java.awt.Rectangle(x-radius, y-radius, 2*radius, 2*radius ) ;
        return s ; }
    
    public void setRadius(double newRadius) {
    	monitor.enter();
    	this.radius = newRadius ;
    	monitor.leave() ;
    }
    
    public double getRadius() {
    	monitor.enter();
    	double r = radius ;
    	monitor.leave() ;
    	return r ; }

    public Shape getRadarShape() {
        /** @todo This should really be protected by the monitor,
         * as it accesses the state. But I can't do that
         * until the monitors are made reentrant */
        int x = (int) state.position.x ;
        int y = (int) state.position.y ;
        double degrees =  -(state.orientation + radarDirection);
        Shape s = new Arc2D.Double(x-radarRadius, y-radarRadius,
                                   2*radarRadius, 2*radarRadius,
                                   degrees-radarBeamAngle/2.0, radarBeamAngle,
                                   Arc2D.PIE ) ;
        return s ; }

    public void setRadarRange( double range ) {
        radarRadius = range ; }

    public void setRadarDirection( double direction ) {
        radarDirection = direction ; }

    public void setRadarBeamAngle( double beamAngle ) {
        radarBeamAngle = beamAngle ; }

    public boolean checkRadar() {
        Shape radar = getRadarShape() ;
        for( Grob next : getArena() ) {
            if( next != this && next.intersects(radar) ) {
                return true ; } }
        return false ;
    }
    
    public boolean checkRadarFor( Grob thing ) {
        Shape radar = getRadarShape() ;
        return thing != this && thing.intersects( radar ) ;
    }

    public void penDown() {
        monitor.enter();
        try {
            if( lastSegment == null ) {
                Point2D.Double startPoint = (Point2D.Double) state.position.clone() ;
                Point2D.Double endPoint = (Point2D.Double) state.position.clone() ;
                lastSegment = new Segment( startPoint, endPoint, state.orientation, fillColor ) ;
                trail.add(lastSegment); } }
        finally { monitor.leave() ; } }

    public void penUp() {
        monitor.enter();
        try {
            if( lastSegment != null  ) {
                lastSegment = null ; } }
        finally { monitor.leave() ; } }

    public void clearTrail() {
        monitor.enter();
        try {
            if( lastSegment != null ) {
                lastSegment = null ; }
            trail = new LinkedList<Segment>() ;
            trailLength = 0.0 ; }
        finally { monitor.leave() ; } }

    public void setTrailLength( double newLength ) {
        monitor.enter();
        try {
            maxTrailLength = newLength ;  }
        finally { monitor.leave() ; } }

    public double getTrailLength( ) {
        monitor.enter();
        try {
            return maxTrailLength ;  }
        finally { monitor.leave() ; } }

    public boolean doesTrailIntersectTrail( Turtle otherTurtle ) {
        monitor.enter() ;
        try {
            for( int i = 0 ; i < trail.size() ; ++i ) {
                int j ;
                if( otherTurtle == this ) {
                    j = i+2 ; }
                else {
                    j = 0 ; }
                for(  ; j < trail.size() ; ++j) {
                    if( intersect( trail.get(i), otherTurtle.trail.get(j) ) ) {
                        return true ; } } }
            return false ; }
        finally { monitor.leave() ; } }

    protected void updateHook( State newState ) {
        if( lastSegment != null ) {
            if( (int)lastSegment.endPoint.x != (int)newState.position.x
             || (int)lastSegment.endPoint.y != (int)newState.position.y ) {
                // penIsDown and movement is significant.
                if( lastSegment.orientation == newState.orientation
                && colinear( lastSegment.startPoint, lastSegment.endPoint, newState.position ) ) {
                    // No change of orientation and the new point is on the extension
                    // of the last segment, so extend the last segment
                    double oldLastSegLength = lastSegment.length() ;
                    lastSegment.endPoint.x = newState.position.x ;
                    lastSegment.endPoint.y = newState.position.y ;
                    double newLastSegLength = lastSegment.length() ;
                    trailLength += newLastSegLength - oldLastSegLength ; }
                else {
                    // Start a new segment
                    Point2D.Double startPoint = lastSegment.endPoint ;
                    Point2D.Double endPoint = (Point2D.Double) newState.position.clone() ;
                    lastSegment = new Segment( startPoint, endPoint, newState.orientation,  fillColor) ;
                    double newLastSegLength = lastSegment.length() ;
                    trail.add(lastSegment);
                    trailLength += newLastSegLength ; }
                
//                System.out.println( "Trail length = "+ trailLength ) ;

                // Trim the trail ;
                double amountToChop = trailLength - maxTrailLength ;
//                System.out.println( "amountToChop = " + amountToChop ) ;
                while(amountToChop > 0 && trail.size() > 0 ) {
                    Segment firstSegment = trail.get(0) ;
                    double firstSegmentLength = firstSegment.length() ;
//                    System.out.println( "firstSegmentLength = " + firstSegmentLength ) ;
                    if( firstSegmentLength < amountToChop && trail.size() > 1) {
                        trail.remove(0) ;
                        trailLength -= firstSegmentLength ;
                        amountToChop -= firstSegmentLength ; }
                    else {
                        double ratio = amountToChop / firstSegmentLength ;
//                        System.out.println( "ratio = " + ratio ) ;
//                        System.out.println( "firstSegment.startPoint.x = " + firstSegment.startPoint.x ) ;
//                        System.out.println( "firstSegment.endPoint.x = " + firstSegment.endPoint.x ) ;
//                        System.out.println( "firstSegment.startPoint.y = " + firstSegment.startPoint.y ) ;
//                        System.out.println( "firstSegment.endPoint.y = " + firstSegment.endPoint.y ) ;
//                        System.out.println( "ajusted x = " + (firstSegment.startPoint.x + ratio * (firstSegment.endPoint.x - firstSegment.startPoint.x)) ) ;
//                        System.out.println( "ajusted y = " + (firstSegment.startPoint.y + ratio * (firstSegment.endPoint.y - firstSegment.startPoint.y)) ) ;
                        firstSegment.startPoint.x += ratio * (firstSegment.endPoint.x - firstSegment.startPoint.x) ;
                        firstSegment.startPoint.y += ratio * (firstSegment.endPoint.y - firstSegment.startPoint.y) ;
                        double newFirstSegmentLength = firstSegment.length() ;
                        trailLength += newFirstSegmentLength - firstSegmentLength ;
                        amountToChop = 0.0 ; } }
            }
        }
    }

    /** Return true if p is roughly on the segment line between p0 and p1. */
    private static boolean colinear( Point2D.Double p0, Point2D.Double p, Point2D.Double p1 ) {
        double dx = p1.x - p0.x ;
        double dy = p1.y - p0.y ;
        if( -FUZZ <= dx && dx < FUZZ ) {
            // Line is roughly vertical
            return p0.y <= p.y && p.y <= p1.y || p1.y <= p.y && p.y <= p0.y ; }
        else if( -FUZZ <= dy && dy < FUZZ ) {
            // Line is roughly horizontal
            return p0.x <= p.x && p.x <= p1.x || p1.x <= p.x && p.x <= p0.x ; }
        else {
            // Line is on a slant.
        	// TODO: Check if this algorithm is right.
        	// Alternative approach. See if d(p0,p)+d(p,p1) ~= d(p0,p1).
            double sx = (p.x - p0.x) / dx ;
            double sy = (p.y - p0.y) / dy ;
            if( 0 <= sx && sx <= 1.0 && 0 <= sy && sy <= 1.0 ) {
                double ds = sx-sy ;
                return -FUZZ <= ds && ds <= FUZZ ; }
            else {
                return false ; } } }

    static boolean oldIntersect(Segment m, Segment n ) {
        double a = m.startPoint.x ;
        double b = m.endPoint.x ;
        double c = n.startPoint.x ;
        double d = n.endPoint.x ;
        double e = m.startPoint.y ;
        double f = m.endPoint.y ;
        double g = n.startPoint.y ;
        double h = n.endPoint.y ;
        double denom = -f*d+f*c+e*d-e*c+h*b-h*a-g*b+g*a ;
        if( -FUZZ < denom && denom < +FUZZ ) {
            return colinear( m.startPoint, n.startPoint, m.endPoint )
                || colinear( m.startPoint, n.endPoint, m.endPoint )
                || colinear( n.startPoint, m.startPoint, n.endPoint )
                || colinear( n.startPoint, m.endPoint, n.endPoint ) ;
        }
        else {
            double t = (-h*a+g*a-d*g+c*h+e*d-e*c) / denom ;
            double s = (-e*b+f*a-f*c+e*c+g*b-g*a) / denom ;
            return 0 <= t && t <= 1.0 && 0 <= s && s <= 1.0 ; }
    }

    static boolean intersect(Segment m, Segment n ) {
        Point2D.Double p0 = m.startPoint ;
        Point2D.Double p1 = m.endPoint ;
        Point2D.Double p2 = n.startPoint ;
        Point2D.Double p3 = n.endPoint ;
//        System.out.print("Intersect of ");
//        System.out.print("("+p0.x +","+p0.y+")" );
//        System.out.print("("+p1.x +","+p1.y+")" );
//        System.out.print(" and ");
//        System.out.print("("+p2.x +","+p2.y+")" );
//        System.out.print("("+p3.x +","+p3.y+")" );
//        System.out.println();
        boolean value ;
        int a = side( p0, p1, p2 ) ;
        int b = side( p0, p1, p3 ) ;
        if( a==0 || b==0 || a == -b ) {
            int c = side( p2, p3, p0 ) ;
            int d = side( p2, p3, p1 ) ;
//            System.out.println("a ="+a+" "+"b ="+b+" "+"c ="+c+" "+"d ="+d);
            value = c==0 || d == 0 || c == -d ; }
        else {
            value = false ; }
//        System.out.println("is "+value);
        return value ;
    }

    /** Return 0 if p is on line (p0, p1), -1 if it is on one side and +1 if
     * it is on the other. Which side is which doesn't matter.
     * */
    static int side( Point2D.Double p0, Point2D.Double p1, Point2D.Double p ) {
        // Translate to the origin
        p1 = new Point2D.Double( p1.x - p0.x, p1.y - p0.y ) ;
        p = new Point2D.Double( p.x - p0.x, p.y - p0.y ) ;

        // Flip the line to the first quadrant or the +ive axes
        if( p1.y < 0.0 ) { p1.y = - p1.y ; p.y = - p.y ; }
        if( p1.x < 0.0 ) { p1.x = -p1.x ; p.x = - p.x ; }

        if( fuzzyEqual( p1.x, 0.0) ) {
            // The line is the positive y-axis
//            System.out.println("Case 0");
            return fuzzyEqual( p.x, 0.0 ) ? 0
                 : p.x < 0.0              ? +1
                 :                          -1 ; }
        else {
            // The line is in the first quadrant or the +ive x axis.
            if( fuzzyEqual( p.x, 0.0 ) ) {
//                System.out.println("Case 1");
                // Point is on the y-axis
                return fuzzyEqual( p.y, 0.0 ) ? 0
                     : p.y < 0.0               ? +1
                     :                          -1 ; }
            else if( p.x > 0.0 ) {
//                System.out.println("Case 2");
                // Point is left of the y-axis
                double lineSlope = p1.y / p1.x ;
                double pointSlope = p.y / p.x ;
                return fuzzyEqual( pointSlope, lineSlope ) ? 0
                     : pointSlope < lineSlope              ? +1
                     :                                       -1 ; }
            else {
//                System.out.println("Case 3");
               // Point is to the right of the y-axis
                double lineSlope = p1.y / p1.x ;
                double pointSlope = p.y / p.x ;
                return fuzzyEqual( pointSlope, lineSlope ) ? 0
                     : pointSlope < lineSlope              ? -1
                     :                                       +1 ; }
        }
    }

    static boolean fuzzyEqual( double x, double y ) {
        return x-FUZZ <= y && y <= x+FUZZ ; }


}

class Segment {
    Point2D.Double startPoint, endPoint ;
    double orientation ;
    Color color ;

    Segment( Point2D.Double start, Point2D.Double end, double orientn, Color col ) {
        startPoint = start ;
        endPoint = end ;
        orientation = orientn ;
        color = col ; }

    double length() {
        return startPoint.distance( endPoint ) ; }
}